import h5py

from .common import TestCase, make_name


class TestCompletions(TestCase):

    def test_root_group_completions(self):
        # Test completions on top-level file.
        with h5py.File(self.mktemp(), 'w') as f:
            f.create_group('h')
            f.create_group('g')
            f.create_dataset('data', [1, 2, 3], "f4")
            self.assertEqual(
            f._ipython_key_completions_(),
                ['data', 'g', 'h'],
            )

            f.create_group('data2')
            # Test that order is alphabetical, and that there is no
            # internal ordering between groups and datasets.
            self.assertEqual(
                f._ipython_key_completions_(),
                ['data', 'data2', 'g', 'h'],
            )

    def test_subgroup_completions(self):
        g = self.f.create_group(make_name())
        g.create_dataset('g_data2', [4, 5, 6], "f4")
        g.create_dataset('g_data1', [1, 2, 3], "f4")
        self.assertEqual(
            g._ipython_key_completions_(),
            ['g_data1', 'g_data2'],  # Order is alphabetical
        )

        g.create_dataset('g_data3', [7, 8, 9], "f")
        self.assertEqual(
            g._ipython_key_completions_(),
            ['g_data1', 'g_data2', 'g_data3'],
        )

    def test_attrs_completions(self):
        # In pytest-run-parallel, let every thread run on a different set of attrs
        attrs = self.f.create_group(make_name()).attrs

        # Write out of alphabetical order to test that completions come back in
        # alphabetical order, as opposed to, say, insertion order.
        attrs['b'] = 1
        attrs['a'] = 2
        self.assertEqual(
            attrs._ipython_key_completions_(),
            ['a', 'b']
        )

        attrs['c'] = 3
        self.assertEqual(
            attrs._ipython_key_completions_(),
            ['a', 'b', 'c']
        )



# This file is part of h5py, a Python interface to the HDF5 library.
#
# http://www.h5py.org
#
# Copyright 2008-2013 Andrew Collette and contributors
#
# License:  Standard 3-clause BSD; see "license.txt" for full license terms
#           and contributor agreement.

"""
Testing native complex number datatypes.
"""

import sys

import numpy as np
import pytest

import h5py
from h5py.h5 import get_config  # type: ignore
from h5py import h5t
from .common import make_name
from .data_files import get_data_file_path

cfg = get_config()

pytestmark = [
    pytest.mark.skipif(
        h5py.version.hdf5_version_tuple < (2, 0, 0),
        reason="Requires HDF5 >= 2.0",
    ),
    pytest.mark.skipif(
        not cfg.has_native_complex,
        reason="Native HDF5 complex number datatypes not available",
    ),
]

REQUIRE_NUMPY_COMPLEX256 = pytest.mark.skipif(
    not hasattr(np, "complex256"),
    reason="complex256 type is not available in numpy",
)

NATIVE_BYTE_ORDER = h5t.ORDER_BE if sys.byteorder == "big" else h5t.ORDER_LE


@pytest.mark.parametrize(
    "h5type_str, size, order",
    [
        pytest.param("COMPLEX_IEEE_F16LE", 4, h5t.ORDER_LE),
        pytest.param("COMPLEX_IEEE_F16BE", 4, h5t.ORDER_BE, id="F16BE"),
        pytest.param("COMPLEX_IEEE_F32LE", 8, h5t.ORDER_LE, id="F32LE"),
        pytest.param("COMPLEX_IEEE_F32BE", 8, h5t.ORDER_BE, id="F32BE"),
        pytest.param("COMPLEX_IEEE_F64LE", 16, h5t.ORDER_LE, id="F64LE"),
        pytest.param("COMPLEX_IEEE_F64BE", 16, h5t.ORDER_BE, id="F64BE"),
        pytest.param("NATIVE_FLOAT_COMPLEX", 8, NATIVE_BYTE_ORDER, id="float-native"),
        pytest.param("NATIVE_DOUBLE_COMPLEX", 16, NATIVE_BYTE_ORDER, id="double-native"),
        pytest.param(
            "NATIVE_LDOUBLE_COMPLEX", 32, NATIVE_BYTE_ORDER,
            id="long-native",
            marks=REQUIRE_NUMPY_COMPLEX256,
        )
    ]
)
def test_hdf5_dtype(h5type_str, size, order):
    """Low-level checks of HDF5 native complex number datatypes"""
    h5type = getattr(h5t, h5type_str)
    assert isinstance(h5type, h5t.TypeComplexID)
    assert h5type.get_size() == size
    assert h5type.get_order() == order

H5_VS_NUMPY_DTYPES = [
    pytest.param("COMPLEX_IEEE_F32LE", "<c8", id="<c8"),
    pytest.param("COMPLEX_IEEE_F32BE", ">c8", id=">c8"),
    pytest.param("COMPLEX_IEEE_F64LE", "<c16", id="<c16"),
    pytest.param("COMPLEX_IEEE_F64BE", ">c16", id=">c16"),
    pytest.param("NATIVE_FLOAT_COMPLEX", "=c8", id="=c8"),
    pytest.param("NATIVE_DOUBLE_COMPLEX", "=c16", id="=c16"),
    pytest.param(
        "NATIVE_LDOUBLE_COMPLEX", "=c32", id="=c32",
        marks=REQUIRE_NUMPY_COMPLEX256,
    ),
]

@pytest.mark.parametrize("h5type_str, dt", H5_VS_NUMPY_DTYPES)
def test_cmplx_type_trnslt(h5type_str, dt):
    """Translate native HDF5 complex number datatype to NumPy dtype"""
    assert getattr(h5t, h5type_str).dtype == np.dtype(dt)


def check_compound_complex_datatype(datatype, np_dtype):
    """Check h5py's old complex number type"""
    assert isinstance(datatype, h5t.TypeCompoundID)
    assert datatype.get_nmembers() == 2
    assert datatype.get_member_name(0) == cfg._r_name
    assert datatype.get_member_name(1) == cfg._i_name
    assert isinstance(datatype.get_member_type(0), h5t.TypeFloatID)
    assert isinstance(datatype.get_member_type(1), h5t.TypeFloatID)

    assert (
       datatype.get_member_type(0).get_size()
       + datatype.get_member_type(1).get_size()
    ) == np_dtype.itemsize

    match (np_dtype.byteorder, sys.byteorder):
        case (">", _) | ("=", "big"):
            expected = h5t.ORDER_BE
        case ("<", _) | ("=", "little"):
            expected = h5t.ORDER_LE
        case _ as _unreachable:
            raise AssertionError

    assert datatype.get_member_type(0).get_order() == expected
    assert datatype.get_member_type(1).get_order() == expected


@pytest.mark.parametrize("dt", ["<c8", ">c8", "<c16", ">c16"])
def test_default_create(writable_file, dt):
    """Test default translation when creating datasets & attributes"""
    complex_array = (np.random.rand(100) + 1j * np.random.rand(100)).astype(dt)
    ds = writable_file.create_dataset(make_name(dt), data=complex_array)
    c = np.array(1.9 + 1j * 6.7, dtype=dt)
    ds.attrs["c"] = c
    check_compound_complex_datatype(ds.id.get_type(), np.dtype(dt))
    check_compound_complex_datatype(ds.attrs.get_id("c").get_type(), np.dtype(dt))
    np.testing.assert_array_equal(ds[...], complex_array)
    np.testing.assert_array_equal(ds.attrs["c"], c)


@pytest.mark.parametrize("h5type_str, dt", H5_VS_NUMPY_DTYPES)
def test_explicit_create(writable_file, h5type_str, dt):
    """Explicitly use native complex datatype to create datasets & attributes"""
    h5type = getattr(h5t, h5type_str)
    np_dt = np.dtype(dt)
    complex_array = (np.random.rand(100) + 1j * np.random.rand(100)).astype(np_dt)
    ds = writable_file.create_dataset(make_name(dt), (100,), dtype=h5py.Datatype(h5type))
    ds[:] = complex_array
    c = np.array(1.9 + 1j * 6.7, dtype=np_dt)
    ds.attrs.create("c", c, dtype=h5py.Datatype(h5type))
    assert isinstance(ds.id.get_type(), h5t.TypeComplexID)
    assert isinstance(ds.attrs.get_id("c").get_type(), h5t.TypeComplexID)
    np.testing.assert_array_equal(ds[...], complex_array)
    np.testing.assert_array_equal(ds.attrs["c"], c)


def test_dtype_cmpnd_cmplx():
    """Check the resulting dtype of the two-field compound data as complex numbers"""
    with h5py.File(get_data_file_path("compound-dtype-complex.h5"), mode="r") as f:
        for obj in filter(lambda obj: isinstance(obj, h5py.Dataset), f.values()):
            check_compound_complex_datatype(obj.id.get_type(), obj.dtype)



# This file is part of h5py, a Python interface to the HDF5 library.
#
# http://www.h5py.org
#
# Copyright 2008-2013 Andrew Collette and contributors
#
# License:  Standard 3-clause BSD; see "license.txt" for full license terms
#           and contributor agreement.

"""
    Tests the h5py.AttributeManager.create() method.
"""

import numpy as np
from .. import h5t, h5a

from .common import TestCase, make_name

class TestArray(TestCase):

    """
        Check that top-level array types can be created and read.
    """

    def test_int(self):
        # See issue 498
        name = make_name()
        dt = np.dtype('(3,)i')
        data = np.arange(3, dtype='i')

        self.f.attrs.create(name, data=data, dtype=dt)

        aid = h5a.open(self.f.id, name.encode('utf-8'))

        htype = aid.get_type()
        self.assertEqual(htype.get_class(), h5t.ARRAY)

        out = self.f.attrs[name]

        self.assertArrayEqual(out, data)

    def test_string_dtype(self):
        # See issue 498 discussion

        self.f.attrs.create(make_name(), data=42, dtype='i8')

    def test_str(self):
        # See issue 1057
        name = make_name()
        self.f.attrs.create(name, chr(0x03A9))
        out = self.f.attrs[name]
        self.assertEqual(out, chr(0x03A9))
        self.assertIsInstance(out, str)

    def test_tuple_of_unicode(self):
        # Test that a tuple of unicode strings can be set as an attribute. It will
        # be converted to a numpy array of vlen unicode type:
        name = make_name()
        data = ('a', 'b')
        self.f.attrs.create(name, data=data)
        result = self.f.attrs[name]
        self.assertTrue(all(result == data))
        self.assertEqual(result.dtype, np.dtype('O'))

    def test_unicode_np_array(self):
        # However, a numpy array of type U being passed in will not be
        # automatically converted, and should raise an error as it does
        # not map to a h5py dtype
        data = np.array(['a', 'b'], dtype='U1')
        with self.assertRaises(TypeError):
            self.f.attrs.create('x', data=data)

    def test_shape_scalar(self):
        name = make_name()
        self.f.attrs.create(name, data=42, shape=1)
        result = self.f.attrs[name]
        self.assertEqual(result.shape, (1,))

    def test_shape_array(self):
        name = make_name()
        self.f.attrs.create(name, data=np.arange(3), shape=3)
        result = self.f.attrs[name]
        self.assertEqual(result.shape, (3,))

    def test_dtype(self):
        dt = np.dtype('(3,)i')
        array = np.arange(3, dtype='i')
        self.f.attrs.create(make_name(), data=array, dtype=dt)
        # Array dtype shape is incompatible with data shape
        array = np.arange(4, dtype='i')
        with self.assertRaises(ValueError):
            self.f.attrs.create('x', data=array, dtype=dt)
        # Shape of new attribute conflicts with shape of data
        dt = np.dtype('()i')
        with self.assertRaises(ValueError):
            self.f.attrs.create('x', data=array, shape=(5,), dtype=dt)

    def test_key_type(self):
        with self.assertRaises(TypeError):
            self.f.attrs.create(1, data=('a', 'b'))



import numpy as np

from h5py import h5s
from h5py._selector import Selector

class Helper:
    def __init__(self, shape: tuple):
        self.shape = shape

    def __getitem__(self, item) -> h5s.SpaceID:
        if not isinstance(item, tuple):
            item = (item,)
        space = h5s.create_simple(self.shape)
        sel = Selector(space)
        sel.make_selection(item)
        return space


def test_same_shape():
    s1 = Helper((5, 6))[:3, :4]
    s2 = Helper((5, 6))[2:, 2:]
    assert s1.select_shape_same(s2)

    s3 = Helper((5, 6))[:4, :3]
    assert not s1.select_shape_same(s3)

def test_select_copy():
    s1 = h5s.create_simple((50,))
    s1.select_hyperslab((5,), (1,), block=(5,), op=h5s.SELECT_SET)
    s2 = h5s.create_simple((50,))
    s2.select_hyperslab((23,), (1,), block=(3,), op=h5s.SELECT_SET)
    s2.select_copy(s1)
    np.testing.assert_array_equal(
        s1.get_select_hyper_blocklist(), s2.get_select_hyper_blocklist()
    )

def test_combine_select():
    # combine_select should create a new space without modifying the original
    s1 = h5s.create_simple((50,))
    s1.select_hyperslab((5,), (1,), block=(5,), op=h5s.SELECT_SET)

    s2 = h5s.create_simple((50,))
    s2.select_hyperslab((20,), (1,), block=(2,), op=h5s.SELECT_SET)

    s3 = s1.combine_select(s2, op=h5s.SELECT_OR)
    s2.select_hyperslab((5,), (1,), block=(5,), op=h5s.SELECT_OR)
    np.testing.assert_array_equal(
        s3.get_select_hyper_blocklist(), s2.get_select_hyper_blocklist()
    )
    assert not np.array_equal(
        s3.get_select_hyper_blocklist(), s1.get_select_hyper_blocklist()
    )

def test_modify_select():
    # combine_select should create a new space without modifying the original
    s1 = h5s.create_simple((50,))
    s1.select_hyperslab((5,), (1,), block=(5,), op=h5s.SELECT_SET)

    s2 = h5s.create_simple((50,))
    s2.select_hyperslab((20,), (1,), block=(2,), op=h5s.SELECT_SET)

    s1.modify_select(s2, op=h5s.SELECT_OR)
    s2.select_hyperslab((5,), (1,), block=(5,), op=h5s.SELECT_OR)
    np.testing.assert_array_equal(
        s1.get_select_hyper_blocklist(), s2.get_select_hyper_blocklist()
    )



# This file is part of h5py, a Python interface to the HDF5 library.
#
# http://www.h5py.org
#
# Copyright 2008-2013 Andrew Collette and contributors
#
# License:  Standard 3-clause BSD; see "license.txt" for full license terms
#           and contributor agreement.
import os
import threading
import time

import pytest

from h5py import _objects as o
from .common import TestCase


class TestObjects(TestCase):

    def test_invalid(self):
        # Check for segfault on close
        oid = o.ObjectID(0)
        del oid
        oid = o.ObjectID(1)
        del oid

    def test_equality(self):
        # Identifier-based equality
        oid1 = o.ObjectID(42)
        oid2 = o.ObjectID(42)
        oid3 = o.ObjectID(43)

        self.assertEqual(oid1, oid2)
        self.assertNotEqual(oid1, oid3)

    def test_hash(self):
        # Default objects are not hashable
        oid = o.ObjectID(42)
        with self.assertRaises(TypeError):
            hash(oid)

    @pytest.mark.thread_unsafe(reason="fork() from a thread may deadlock")
    @pytest.mark.filterwarnings(
        # https://github.com/python/cpython/pull/100229
        r"ignore:.*use of fork\(\) may lead to deadlocks:DeprecationWarning"
    )
    @pytest.mark.skipif(not hasattr(os, "fork"), reason="fork() not available")
    def test_phil_fork_with_threads(self):
        """Test that handling of the phil Lock after fork is correct.

        h5py uses os.register_at_fork() to cause os.fork() to acquire the phil lock
        before forking and release it afterwards, so that the global state of libhdf5
        cannot be cloned in a corrupted state.
        """
        thread_acquired_phil_event = threading.Event()

        def f():
            # Simulate another thread holding the phil lock while it updates
            # the libhdf5 global state
            with o.phil:
                thread_acquired_phil_event.set()
                time.sleep(1)

        thread = threading.Thread(target=f)
        thread.start()
        thread_acquired_phil_event.wait()

        try:
            # Now fork the current (main) thread while the other thread holds the lock.
            # os.fork() acquires the phil lock, so this will block until the other
            # thread releases it.
            pid = os.fork()
            if pid == 0:
                # Child process
                # If we handle the phil lock correctly, this should not deadlock, and we
                # should be able to acquire the lock here.
                if o.phil.acquire(blocking=False):
                    o.phil.release()
                    os._exit(0)
                else:
                    os._exit(1)
            else:
                # Parent process
                assert o.phil.acquire(blocking=False)
                o.phil.release()
                # Wait for the child process to finish
                _, status = os.waitpid(pid, 0)
                assert os.WIFEXITED(status)
                assert os.WEXITSTATUS(status) == 0
        finally:
            thread.join()



# This file is part of h5py, a Python interface to the HDF5 library.
#
# http://www.h5py.org
#
# Copyright 2008-2013 Andrew Collette and contributors
#
# License:  Standard 3-clause BSD; see "license.txt" for full license terms
#           and contributor agreement.

"""
    Read-only S3 virtual file driver (VFD) test module.
"""

import h5py
from h5py._hl.files import make_fapl
import pytest


pytestmark = [
    pytest.mark.skipif(
        not h5py.h5.get_config().ros3,
        reason="ros3 driver not available")
]


@pytest.mark.network
@pytest.mark.parametrize(
    "kwargs",
    [
        pytest.param(
            {},
            id="HDF5-v1",
            marks=pytest.mark.skipif(
                h5py.version.hdf5_version_tuple >= (2, 0, 0),
                reason="Requires HDF5 < 2.0",
            ),
        ),
        pytest.param(
            {"aws_region": b"us-east-2"},
            id="HDF5-v2",
            marks=pytest.mark.skipif(
                h5py.version.hdf5_version_tuple < (2, 0, 0),
                reason="Requires HDF5 >= 2.0",
            ),
        ),
    ],
)
def test_ros3(kwargs):
    """ ROS3 driver and options """

    with h5py.File("https://dandiarchive.s3.amazonaws.com/ros3test.hdf5", 'r',
                   driver='ros3', **kwargs) as f:
        assert f
        assert 'mydataset' in f.keys()
        assert f["mydataset"].shape == (100,)


@pytest.mark.parametrize(
    "exc,match_exc",
    [
        pytest.param(
            ValueError,
            [
                "AWS region required for s3:// location",
                r"^foo://wrong/scheme: S3 location must begin with",
            ],
            id="HDF5-v1",
            marks=pytest.mark.skipif(
                h5py.version.hdf5_version_tuple >= (2, 0, 0),
                reason="Requires HDF5 < 2.0",
            ),
        ),
        pytest.param(
            OSError,
            [None, "can't parse object key from path"],
            id="HDF5-v2",
            marks=pytest.mark.skipif(
                h5py.version.hdf5_version_tuple < (2, 0, 0),
                reason="Requires HDF5 >= 2.0",
            ),
        ),
    ],
)
def test_ros3_s3_fails(exc, match_exc):
    """ROS3 exceptions for s3:// location"""
    with pytest.raises(exc, match=match_exc[0]):
        h5py.File('s3://fakebucket/fakekey', 'r', driver='ros3')

    with pytest.raises(exc, match=match_exc[1]):
        h5py.File('foo://wrong/scheme', 'r', driver='ros3')


@pytest.mark.network
def test_ros3_s3uri():
    """Use S3 URI with ROS3 driver"""
    with h5py.File('s3://dandiarchive/ros3test.hdf5', 'r', driver='ros3',
                   aws_region=b'us-east-2') as f:
        assert f
        assert 'mydataset' in f.keys()
        assert f["mydataset"].shape == (100,)


@pytest.mark.skipif(h5py.version.hdf5_version_tuple < (1, 14, 2),
                    reason='AWS S3 access token support in HDF5 >= 1.14.2')
def test_ros3_temp_token():
    """Set and get S3 access token"""
    token = b'#0123FakeToken4567/8/9'
    fapl = make_fapl('ros3', libver=None, rdcc_nslots=None, rdcc_nbytes=None,
                     rdcc_w0=None, locking=None, page_buf_size=None, min_meta_keep=None,
                     min_raw_keep=None, alignment_threshold=1, alignment_interval=1,
                     meta_block_size=None, session_token=token)
    assert token, fapl.get_fapl_ros3_token()



# This file is part of h5py, a Python interface to the HDF5 library.
#
# http://www.h5py.org
#
# Copyright 2008-2013 Andrew Collette and contributors
#
# License:  Standard 3-clause BSD; see "license.txt" for full license terms
#           and contributor agreement.
import sys
import shlex
from importlib.util import find_spec
from subprocess import call

def run_tests(args=''):
    if find_spec("pytest") is None:
        print("Tests require pytest, pytest not installed")
        return 1

    cli = [sys.executable, "-m", "pytest", "--pyargs", "h5py"]
    cli.extend(shlex.split(args))
    return call(cli)



# This file is part of h5py, a Python interface to the HDF5 library.
#
# http://www.h5py.org
#
# Copyright 2008-2013 Andrew Collette and contributors
#
# License:  Standard 3-clause BSD; see "license.txt" for full license terms
#           and contributor agreement.

"""
    Tests the h5py.File object.
"""

import h5py
from h5py._hl.files import _drivers
from h5py import File

from .common import ut, TestCase, make_name

import pytest
import io
import tempfile
import os


def nfiles():
    return h5py.h5f.get_obj_count(h5py.h5f.OBJ_ALL, h5py.h5f.OBJ_FILE)

def ngroups():
    return h5py.h5f.get_obj_count(h5py.h5f.OBJ_ALL, h5py.h5f.OBJ_GROUP)


class TestDealloc(TestCase):

    """
        Behavior on object deallocation.  Note most of this behavior is
        delegated to FileID.
    """
    @pytest.mark.thread_unsafe(reason="global object counters")
    def test_autoclose(self):
        """ File objects close automatically when out of scope, but
        other objects remain open. """

        start_nfiles = nfiles()
        start_ngroups = ngroups()

        fname = self.mktemp()
        f = h5py.File(fname, 'w')
        g = f['/']

        self.assertEqual(nfiles(), start_nfiles+1)
        self.assertEqual(ngroups(), start_ngroups+1)

        del f

        self.assertTrue(g)
        self.assertEqual(nfiles(), start_nfiles)
        self.assertEqual(ngroups(), start_ngroups+1)

        f = g.file

        self.assertTrue(f)
        self.assertEqual(nfiles(), start_nfiles+1)
        self.assertEqual(ngroups(), start_ngroups+1)

        del g

        self.assertEqual(nfiles(), start_nfiles+1)
        self.assertEqual(ngroups(), start_ngroups)

        del f

        self.assertEqual(nfiles(), start_nfiles)
        self.assertEqual(ngroups(), start_ngroups)


class TestDriverRegistration(TestCase):
    def test_register_driver(self):
        called_with = [None]

        def set_fapl(plist, *args, **kwargs):
            called_with[0] = args, kwargs
            return _drivers['sec2'](plist)

        name = make_name()
        h5py.register_driver(name, set_fapl)
        self.assertIn(name, h5py.registered_drivers())

        fname = self.mktemp()
        h5py.File(fname, driver=name, driver_arg_0=0, driver_arg_1=1,
                  mode='w')

        self.assertEqual(
            called_with,
            [((), {'driver_arg_0': 0, 'driver_arg_1': 1})],
        )

    def test_unregister_driver(self):
        name = make_name()
        h5py.register_driver(name, lambda plist: None)
        self.assertIn(name, h5py.registered_drivers())

        h5py.unregister_driver(name)
        self.assertNotIn(name, h5py.registered_drivers())

        with self.assertRaises(ValueError) as e:
            fname = self.mktemp()
            h5py.File(fname, driver=name, mode='w')

        self.assertEqual(str(e.exception), f"Unknown driver type '{name}'")


class TestCache(TestCase):
    def setUp(self):
        MiB = 1024 * 1024
        if h5py.version.hdf5_version_tuple < (2, 0, 0):
            self.dflt_chunk_cache = MiB
            self.dflt_chunk_nslots = 521
        else:
            self.dflt_chunk_cache = 8 * MiB
            self.dflt_chunk_nslots = 8191

    def test_defaults(self):
        fname = self.mktemp()
        f = h5py.File(fname, 'w')
        self.assertEqual(list(f.id.get_access_plist().get_cache()),
                         [0, self.dflt_chunk_nslots, self.dflt_chunk_cache, 0.75])

    def test_nbytes(self):
        fname = self.mktemp()
        f = h5py.File(fname, 'w', rdcc_nbytes=1024)
        self.assertEqual(list(f.id.get_access_plist().get_cache()),
                         [0, self.dflt_chunk_nslots, 1024, 0.75])

    def test_nslots(self):
        fname = self.mktemp()
        f = h5py.File(fname, 'w', rdcc_nslots=125)
        self.assertEqual(list(f.id.get_access_plist().get_cache()),
                         [0, 125, self.dflt_chunk_cache, 0.75])

    def test_w0(self):
        fname = self.mktemp()
        f = h5py.File(fname, 'w', rdcc_w0=0.25)
        self.assertEqual(list(f.id.get_access_plist().get_cache()),
                         [0, self.dflt_chunk_nslots, self.dflt_chunk_cache, 0.25])


class TestFileObj(TestCase):

    def check_write(self, fileobj):
        f = h5py.File(fileobj, 'w')
        self.assertEqual(f.driver, 'fileobj')
        self.assertEqual(f.filename, repr(fileobj))
        f.create_dataset('test', data=list(range(12)))
        self.assertEqual(list(f), ['test'])
        self.assertEqual(list(f['test'][:]), list(range(12)))
        f.close()

    def check_read(self, fileobj):
        f = h5py.File(fileobj, 'r')
        self.assertEqual(list(f), ['test'])
        self.assertEqual(list(f['test'][:]), list(range(12)))
        self.assertRaises(Exception, f.create_dataset, 'another.test', data=list(range(3)))
        f.close()

    def test_BytesIO(self):
        with io.BytesIO() as fileobj:
            self.assertEqual(len(fileobj.getvalue()), 0)
            self.check_write(fileobj)
            self.assertGreater(len(fileobj.getvalue()), 0)
            self.check_read(fileobj)

    def test_file(self):
        fname = self.mktemp()
        try:
            with open(fname, 'wb+') as fileobj:
                self.assertEqual(os.path.getsize(fname), 0)
                self.check_write(fileobj)
                self.assertGreater(os.path.getsize(fname), 0)
                self.check_read(fileobj)
            with open(fname, 'rb') as fileobj:
                self.check_read(fileobj)
        finally:
            os.remove(fname)

    @pytest.mark.filterwarnings(
        # at least on Windows and MacOS, a resource warning may be emitted
        # when this test returns
        "ignore::ResourceWarning"
    )
    def test_TemporaryFile(self):
        # in this test, we check explicitly that temp file gets
        # automatically deleted upon h5py.File.close()...
        fileobj = tempfile.NamedTemporaryFile()
        fname = fileobj.name
        f = h5py.File(fileobj, 'w')
        del fileobj
        # ... but in your code feel free to simply
        # f = h5py.File(tempfile.TemporaryFile())

        f.create_dataset('test', data=list(range(12)))
        self.assertEqual(list(f), ['test'])
        self.assertEqual(list(f['test'][:]), list(range(12)))
        self.assertTrue(os.path.isfile(fname))
        f.close()
        self.assertFalse(os.path.isfile(fname))

    def test_exception_open(self):
        self.assertRaises(Exception, h5py.File, None,
                          driver='fileobj', mode='x')
        self.assertRaises(Exception, h5py.File, 'rogue',
                          driver='fileobj', mode='x')
        self.assertRaises(Exception, h5py.File, self,
                          driver='fileobj', mode='x')

    def test_exception_read(self):

        class BrokenBytesIO(io.BytesIO):
            def readinto(self, b):
                raise Exception('I am broken')

        f = h5py.File(BrokenBytesIO(), 'w')
        f.create_dataset('test', data=list(range(12)))
        self.assertRaises(Exception, list, f['test'])

    def test_exception_write(self):

        class BrokenBytesIO(io.BytesIO):
            allow_write = False
            def write(self, b):
                if self.allow_write:
                    return super().write(b)
                else:
                    raise Exception('I am broken')

        bio = BrokenBytesIO()
        f = h5py.File(bio, 'w')
        try:
            self.assertRaises(Exception, f.create_dataset, 'test',
                              data=list(range(12)))
        finally:
            # Un-break writing so we can close: errors while closing get messy.
            bio.allow_write = True
            f.close()

    @ut.skip("Incompletely closed files can cause segfaults")
    def test_exception_close(self):
        fileobj = io.BytesIO()
        f = h5py.File(fileobj, 'w')
        fileobj.close()
        self.assertRaises(Exception, f.close)

    def test_exception_writeonly(self):
        # HDF5 expects read & write access to a file it's writing;
        # check that we get the correct exception on a write-only file object.
        fileobj = open(os.path.join(self.tempdir, make_name("foo{}.h5")), 'wb')
        f = h5py.File(fileobj, 'w')
        group = f.create_group("group")
        with self.assertRaises(io.UnsupportedOperation):
            group.create_dataset("data", data='foo', dtype=h5py.string_dtype())
        f.close()
        fileobj.close()


    def test_method_vanish(self):
        fileobj = io.BytesIO()
        f = h5py.File(fileobj, 'w')
        f.create_dataset('test', data=list(range(12)))
        self.assertEqual(list(f['test'][:]), list(range(12)))
        fileobj.readinto = None
        self.assertRaises(Exception, list, f['test'])


class TestTrackOrder(TestCase):
    def populate(self, f):
        for i in range(100):
            # Mix group and dataset creation.
            if i % 10 == 0:
                f.create_group(str(i))
            else:
                f[str(i)] = [i]

    def test_track_order(self):
        fname = self.mktemp()
        f = h5py.File(fname, 'w', track_order=True)  # creation order
        self.populate(f)
        self.assertEqual(list(f), [str(i) for i in range(100)])
        f.close()

        # Check order tracking after reopening the file
        f2 = h5py.File(fname)
        self.assertEqual(list(f2), [str(i) for i in range(100)])

    def test_no_track_order(self):
        fname = self.mktemp()
        f = h5py.File(fname, 'w', track_order=False)  # name alphanumeric
        self.populate(f)
        self.assertEqual(list(f),
                         sorted([str(i) for i in range(100)]))


class TestFileMetaBlockSize(TestCase):

    """
        Feature: The meta block size can be manipulated, changing how metadata
        is aggregated and the offset of the first dataset.
    """

    def test_file_create_with_meta_block_size_4096(self):
        # Test a large meta block size of 4 kibibytes
        meta_block_size = 4096
        with File(
            self.mktemp(), 'w',
            meta_block_size=meta_block_size,
            libver="latest"
        ) as f:
            f["test"] = 5
            self.assertEqual(f.meta_block_size, meta_block_size)
            # Equality is expected for HDF5 1.10
            self.assertGreaterEqual(f["test"].id.get_offset(), meta_block_size)

    def test_file_create_with_meta_block_size_512(self):
        # Test a small meta block size of 512 bytes
        # The smallest verifiable meta_block_size is 463
        meta_block_size = 512
        libver = "latest"
        with File(
            self.mktemp(), 'w',
            meta_block_size=meta_block_size,
            libver=libver
        ) as f:
            f["test"] = 3
            self.assertEqual(f.meta_block_size, meta_block_size)
            # Equality is expected for HDF5 1.10
            self.assertGreaterEqual(f["test"].id.get_offset(), meta_block_size)
            # Default meta_block_size is 2048. This should fail if meta_block_size is not set.
            self.assertLess(f["test"].id.get_offset(), meta_block_size*2)



# This file is part of h5py, a Python interface to the HDF5 library.
#
# http://www.h5py.org
#
# Copyright 2008-2013 Andrew Collette and contributors
#
# License:  Standard 3-clause BSD; see "license.txt" for full license terms
#           and contributor agreement.

import sys
import os
import shutil
import inspect
import tempfile
import threading
import subprocess
from contextlib import contextmanager
from functools import wraps

import numpy as np
from numpy.lib.recfunctions import repack_fields
import h5py
from h5py._objects import phil
import unittest as ut


# Check if non-ascii filenames are supported
# Evidently this is the most reliable way to check
# See also h5py issue #263 and ipython #466
# To test for this, run the testsuite with LC_ALL=C
try:
    testfile, fname = tempfile.mkstemp(chr(0x03b7))
except UnicodeError:
    UNICODE_FILENAMES = False
else:
    UNICODE_FILENAMES = True
    os.close(testfile)
    os.unlink(fname)
    del fname
    del testfile


class TestCase(ut.TestCase):

    """
        Base class for unit tests.
    """

    @classmethod
    def setUpClass(cls):
        cls.tempdir = tempfile.mkdtemp(prefix='h5py-test_')

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.tempdir)

    def mktemp(self, suffix='.hdf5', prefix='tmp', dir=None):
        if dir is None:
            dir = self.tempdir
        return tempfile.mktemp(suffix, make_name(prefix), dir=dir)

    def mktemp_mpi(self, comm=None, suffix='.hdf5', prefix='', dir=None):
        if comm is None:
            from mpi4py import MPI
            comm = MPI.COMM_WORLD
        fname = None
        if comm.Get_rank() == 0:
            fname = self.mktemp(suffix, prefix, dir)
        fname = comm.bcast(fname, 0)
        return fname

    def setUp(self):
        self.f = h5py.File(self.mktemp(), 'w')

    def tearDown(self):
        try:
            if self.f:
                self.f.close()
        except:
            pass

    def assertSameElements(self, a, b):
        for x in a:
            match = False
            for y in b:
                if x == y:
                    match = True
            if not match:
                raise AssertionError("Item '%s' appears in a but not b" % x)

        for x in b:
            match = False
            for y in a:
                if x == y:
                    match = True
            if not match:
                raise AssertionError("Item '%s' appears in b but not a" % x)

    def assertArrayEqual(self, dset, arr, message=None, precision=None, check_alignment=True):
        """ Make sure dset and arr have the same shape, dtype and contents, to
            within the given precision, optionally ignoring differences in dtype alignment.

            Note that dset may be a NumPy array or an HDF5 dataset.
        """
        if precision is None:
            precision = 1e-5
        if message is None:
            message = ''
        else:
            message = ' (%s)' % message

        if np.isscalar(dset) or np.isscalar(arr):
            assert np.isscalar(dset) and np.isscalar(arr), \
                'Scalar/array mismatch ("%r" vs "%r")%s' % (dset, arr, message)
            dset = np.asarray(dset)
            arr = np.asarray(arr)

        assert dset.shape == arr.shape, \
            "Shape mismatch (%s vs %s)%s" % (dset.shape, arr.shape, message)
        if dset.dtype != arr.dtype:
            if check_alignment:
                normalized_dset_dtype = dset.dtype
                normalized_arr_dtype = arr.dtype
            else:
                normalized_dset_dtype = repack_fields(dset.dtype)
                normalized_arr_dtype = repack_fields(arr.dtype)

            assert normalized_dset_dtype == normalized_arr_dtype, \
                "Dtype mismatch (%s vs %s)%s" % (normalized_dset_dtype, normalized_arr_dtype, message)

            if not check_alignment:
                if normalized_dset_dtype != dset.dtype:
                    dset = repack_fields(np.asarray(dset))
                if normalized_arr_dtype != arr.dtype:
                    arr = repack_fields(np.asarray(arr))

        if arr.dtype.names is not None:
            for n in arr.dtype.names:
                message = '[FIELD %s] %s' % (n, message)
                self.assertArrayEqual(dset[n], arr[n], message=message, precision=precision, check_alignment=check_alignment)
        elif arr.dtype.kind in ('i', 'f'):
            assert np.all(np.abs(dset[...] - arr[...]) < precision), \
                "Arrays differ by more than %.3f%s" % (precision, message)
        elif arr.dtype.kind == 'O':
            for v1, v2 in zip(dset.flat, arr.flat, strict=True):
                self.assertArrayEqual(v1, v2, message=message, precision=precision, check_alignment=check_alignment)
        else:
            assert np.all(dset[...] == arr[...]), \
                "Arrays are not equal (dtype %s) %s" % (arr.dtype.str, message)

    def assertNumpyBehavior(self, dset, arr, s, skip_fast_reader=False):
        """ Apply slicing arguments "s" to both dset and arr.

        Succeeds if the results of the slicing are identical, or the
        exception raised is of the same type for both.

        "arr" must be a Numpy array; "dset" may be a NumPy array or dataset.
        """
        exc = None
        try:
            arr_result = arr[s]
        except Exception as e:
            exc = type(e)

        s_fast = s if isinstance(s, tuple) else (s,)

        if exc is None:
            self.assertArrayEqual(dset[s], arr_result)

            if not skip_fast_reader:
                with phil:
                    self.assertArrayEqual(
                        dset._fast_reader.read(s_fast),
                        arr_result,
                    )
        else:
            with self.assertRaises(exc):
                dset[s]

            if not skip_fast_reader:
                with self.assertRaises(exc), phil:
                    dset._fast_reader.read(s_fast)

NUMPY_RELEASE_VERSION = tuple([int(i) for i in np.__version__.split(".")[0:2]])

@contextmanager
def closed_tempfile(suffix='', text=None):
    """
    Context manager which yields the path to a closed temporary file with the
    suffix `suffix`. The file will be deleted on exiting the context. An
    additional argument `text` can be provided to have the file contain `text`.
    """
    with tempfile.NamedTemporaryFile(
        'w+t', suffix=suffix, delete=False
    ) as test_file:
        file_name = test_file.name
        if text is not None:
            test_file.write(text)
            test_file.flush()
    yield file_name
    shutil.rmtree(file_name, ignore_errors=True)


def insubprocess(f):
    """Runs a test in its own subprocess"""
    @wraps(f)
    def wrapper(request, *args, **kwargs):
        curr_test = inspect.getsourcefile(f) + "::" + request.node.name
        # get block around test name
        insub = "IN_SUBPROCESS_" + curr_test
        for c in "/\\,:.":
            insub = insub.replace(c, "_")
        defined = os.environ.get(insub, None)
        if defined:
            # We're already running in a subprocess
            return f(request, *args, **kwargs)
        else:
            # Spawn a new interpreter and run pytest in it
            env = os.environ.copy()
            env[insub] = '1'
            env.update(getattr(f, 'subproc_env', {}))

            with closed_tempfile() as stdout:
                with open(stdout, 'w+t') as fh:
                    rtn = subprocess.call([sys.executable, '-m', 'pytest', curr_test],
                                          stdout=fh, stderr=fh, env=env)
                with open(stdout, 'rt') as fh:
                    out = fh.read()

            assert rtn == 0, "\n" + out
    return wrapper


def subproc_env(d):
    """Set environment variables for the @insubprocess decorator"""
    def decorator(f):
        f.subproc_env = d
        return f

    return decorator


MAIN_THREAD_ID = threading.get_ident()


def make_name(template_or_prefix: str = "foo", /) -> str:
    """Return a static name, to be used e.g. as dataset name.

    When running in pytest-run-parallel, append a thread ID to the name.
    This allows running tests on shared resources, e.g. two threads can attempt to write
    to separate datasets on the same File at the same time (even though the actual
    writes will be serialized by the `phil` lock).

    Calling this function twice from the same thread will return the same name.

    Parameters
    ----------
    template_or_prefix
        Either a prefix to which potentially append the thread ID, or a template
        containing exactly one "{}" to be replaced with the thread ID.
    """
    tid = threading.get_ident()
    suffix = "" if tid == MAIN_THREAD_ID else f"-{tid}"
    if "{}" in template_or_prefix:
        return template_or_prefix.format(suffix)
    else:
        return template_or_prefix + suffix


def is_main_thread() -> bool:
    """Return True if the test calling this function is being executed
    in the main thread; False otherwise.
    This can be used to detect when a test is running in pytest-run-parallel.
    that spawns multiple separate threads to run the tests.
    """
    tid = threading.get_ident()
    return tid == MAIN_THREAD_ID



import numpy as np
import pytest

import h5py

from .common import make_name

NUMPY_GE2 = int(np.__version__.split(".")[0]) >= 2
pytestmark = pytest.mark.skipif(not NUMPY_GE2, reason="requires numpy >=2.0")


def test_create_with_dtype_T(writable_file):
    name = make_name()
    ds = writable_file.create_dataset(name, shape=(2, 2), dtype="T")
    data = [["foo", "bar"], ["hello world", ""]]
    ds[:] = data
    a = ds.asstr()[:]
    np.testing.assert_array_equal(a, data)

    ds = writable_file[name]
    assert ds.dtype == object
    np.testing.assert_array_equal(ds.asstr()[:], data)

    ds[0, 0] = "baz"
    data[0][0] = "baz"
    a = ds.astype("T")[:]
    assert a.dtype.kind == "T"
    np.testing.assert_array_equal(a, data)

    ds[0, 0] = np.asarray("123", dtype="O")
    data[0][0] = "123"
    np.testing.assert_array_equal(ds.asstr()[:], data)


def test_fromdata(writable_file):
    nx = make_name("x")
    ny = make_name("y")
    nz = make_name("z")
    data = [["foo", "bar"]]
    np_data = np.asarray(data, dtype="T")
    x = writable_file.create_dataset(nx, data=data, dtype="T")
    y = writable_file.create_dataset(ny, data=data, dtype=np.dtypes.StringDType())
    z = writable_file.create_dataset(nz, data=np_data)

    for ds in (x, y, z):
        assert ds.dtype.kind == "O"
        np.testing.assert_array_equal(ds.astype("T")[:], np_data)
    for n in (nx, ny, nz):
        ds = writable_file[n]
        assert ds.dtype == object
        np.testing.assert_array_equal(ds.asstr()[:], data)
        ds = ds.astype("T")
        assert ds.dtype.kind == "T"
        a = ds[:]
        assert a.dtype.kind == "T"
        np.testing.assert_array_equal(a, data)


def test_fixed_to_variable_width(writable_file):
    data = ["foo", "longer than 8 bytes"]
    x = writable_file.create_dataset(
        make_name(), data=data, dtype=h5py.string_dtype(length=20)
    )
    assert x.dtype == "S20"

    # read T <- S
    y = x.astype("T")
    assert y.dtype.kind == "T"
    assert y[:].dtype.kind == "T"
    np.testing.assert_array_equal(y[:], data)

    # write T -> S
    x[0] = np.asarray("1234", dtype="T")
    data[0] = "1234"
    np.testing.assert_array_equal(y[:], data)


def test_fixed_to_variable_width_too_short(writable_file):
    # Note: this test triggers calls to H5Tconvert which are otherwise skipped.

    data = ["foo", "bar"]
    x = writable_file.create_dataset(
        make_name(), data=data, dtype=h5py.string_dtype(length=3)
    )
    assert x.dtype == "S3"

    # write T -> S
    x[0] = np.asarray("1234", dtype="T")
    np.testing.assert_array_equal(x[:], [b"123", b"bar"])


def test_variable_to_fixed_width(writable_file):
    data = ["foo", "longer than 8 bytes"]
    bdata = [b"foo", b"longer than 8 bytes"]
    x = writable_file.create_dataset(make_name(), data=data, dtype="T")

    # read S <- T
    y = x.astype("S20")
    assert y.dtype == "S20"
    assert y[:].dtype == "S20"
    np.testing.assert_array_equal(y[:], bdata)

    y = x.astype("S3")
    assert y.dtype == "S3"
    assert y[:].dtype == "S3"
    np.testing.assert_array_equal(y[:], [b"foo", b"lon"])

    # write S -> T
    x[0] = np.asarray(b"1234", dtype="S5")
    bdata[0] = b"1234"
    np.testing.assert_array_equal(x[:], bdata)


def test_write_object_into_npystrings(writable_file):
    x = writable_file.create_dataset(make_name(), data=["foo"], dtype="T")
    x[0] = np.asarray("1234", dtype="O")
    np.testing.assert_array_equal(x[:], b"1234")


def test_write_npystrings_into_object(writable_file):
    x = writable_file.create_dataset(
        make_name("x"), data=["foo"], dtype=h5py.string_dtype()
    )
    assert x.dtype == object
    x[0] = np.asarray("1234", dtype="T")
    np.testing.assert_array_equal(x[:], b"1234")

    # Test with HDF5 variable-length strings with ASCII character set
    xa = writable_file.create_dataset(
       make_name("xa"), shape=(1,), dtype=h5py.string_dtype('ascii')
    )
    xa[0] = np.asarray("2345", dtype="T")
    np.testing.assert_array_equal(xa[:], b"2345")


def test_fillvalue(writable_file):
    # Create as NpyString dtype
    x = writable_file.create_dataset(
        make_name("x"), shape=(2,), dtype="T", fillvalue="foo"
    )
    assert isinstance(x.fillvalue, bytes)
    assert x.fillvalue == b"foo"
    assert x[0] == b"foo"

    # Create as object dtype
    y = writable_file.create_dataset(
        make_name("y"), shape=(2,), dtype=h5py.string_dtype(), fillvalue=b"foo"
    )
    assert isinstance(y.fillvalue, bytes)
    assert y.fillvalue == b"foo"
    assert y[0] == b"foo"
    # Convert object dtype to NpyString
    y = y.astype("T")
    assert y[0] == "foo"


def test_empty_string(writable_file):
    data = np.array(["", "a", "b"], dtype="T")
    x = writable_file.create_dataset(make_name(), data=data)
    np.testing.assert_array_equal(x[:], [b"", b"a", b"b"])
    np.testing.assert_array_equal(x.astype("T")[:], data)
    data[:2] = ["c", ""]
    x[:2] = data[:2]
    np.testing.assert_array_equal(x[:], [b"c", b"", b"b"])
    np.testing.assert_array_equal(x.astype("T")[:], data)


def test_astype_nonstring(writable_file):
    x = writable_file.create_dataset(make_name(), shape=(2, ), dtype="i8")
    with pytest.raises(TypeError, match="HDF5 string datatype"):
        x.astype("T")


def test_resized_read(writable_file):
    """Read default values created by resize(). This triggers a special case
    where libhdf5 returns a char** containing NULL pointers.
    """
    l = ["string1", "string2", "string3"]
    data = np.array(l, dtype='T')
    d = writable_file.create_dataset(make_name(), data=data, maxshape=(None,))
    d.resize((10,))

    np.testing.assert_array_equal(d[:], np.array(
        [s.encode() for s in l] + [b''] * 7, dtype=object
    ))
    np.testing.assert_array_equal(d.astype('T')[:], np.array(l + [''] * 7, dtype='T'))
